import json
from gear.utils.keystore import (
    decode_keystore_json,
    priv_to_addr,
)


class account(object):
    def __init__(self):
        self.accounts = {}

    def get_accounts(self):
        return list(self.accounts.keys())

    def get_priv_by_addr(self, adrr):
        return self.accounts[adrr]


class solo(account):
    def __init__(self):
        priv_keys = [
            "dce1443bd2ef0c2631adc1c67e5c93f13dc23a41c18b536effbbdcbcdb96fb65",
            "321d6443bc6177273b5abf54210fe806d451d6b7973bccc2384ef78bbcd0bf51",
            "2d7c882bad2a01105e36dda3646693bc1aaaa45b0ed63fb0ce23c060294f3af2",
            "593537225b037191d322c3b1df585fb1e5100811b71a6f7fc7e29cca1333483e",
            "ca7b25fc980c759df5f3ce17a3d881d6e19a38e651fc4315fc08917edab41058",
            "88d2d80b12b92feaa0da6d62309463d20408157723f2d7e799b6a74ead9a673b",
            "fbb9e7ba5fe9969a71c6599052237b91adeb1e5fc0c96727b66e56ff5d02f9d0",
            "547fb081e73dc2e22b4aae5c60e2970b008ac4fc3073aebc27d41ace9c4f53e9",
            "c8c53657e41a8d669349fc287f57457bd746cb1fcfc38cf94d235deb2cfca81b",
            "87e0eba9c86c494d98353800571089f316740b0cb84c9a7cdf2fe5c9997c7966",
        ]
        self.accounts = {
            priv_to_addr(priv_key): priv_key
            for priv_key in priv_keys
        }


class keystore(account):
    def __init__(self, keystore_path, passcode):
        jsondata = json.loads(open(keystore_path).read())
        priv_key = decode_keystore_json(jsondata, passcode)
        self.accounts = {
            priv_to_addr(priv_key): priv_key,
        }



class Singleton(type):
    def __init__(self, *args, **kwargs):
        self.__instance = None
        super().__init__(*args, **kwargs)

    def __call__(self, *args, **kwargs):
        if self.__instance is None:
            self.__instance = super().__call__(*args, **kwargs)
            return self.__instance
        else:
            return self.__instance



from jsonrpcserver import async_dispatch
import json
from aiohttp import web
from .rpc import make_version
from .meter.account import (
    solo,
    keystore as _keystore,
)
from .meter.client import meter
import requests
import click
from datetime import datetime


res_headers = {
    "Access-Control-Allow-Headers": "*",
    "Access-Control-Allow-Origin": "*",
    "Connection": "keep-alive",
}


async def handle(request, logging=False, debug=False):
    jreq = await request.json()
    reqStr = json.dumps(jreq)
    arrayNeeded = True
    if not isinstance(jreq, list):
        jreq = [jreq]
        arrayNeeded = False

    print('*'*40+"\nRaw Req:", reqStr, "\n"+"*"*40)
    responses = []
    for r in jreq:
        method = r['method']
        # request = await request.text()
        response = await async_dispatch(json.dumps(r), basic_logging=logging, debug=debug)
        if response.wanted:
            print(json.dumps(response.deserialized()))
            responses.append(json.loads(json.dumps(response.deserialized())))

    if len(responses):
        print("-"*40+"\nRPC Call:", method, "\nTime: ", datetime.now().timestamp(),
              "\nRequest:", reqStr, "\nResponse:", json.dumps(responses)+"\n"+"-"*40)
        if arrayNeeded:
            return web.json_response(responses, headers=res_headers, status=response.http_status)
        else:
            return web.json_response(responses[0], headers=res_headers, status=response.http_status)
    else:
        return web.Response(headers=res_headers, content_type="text/plain")


@click.command()
@click.option(
    "--host",
    default="127.0.0.1",
)
@click.option(
    "--port",
    default=8545,
    type=int,
)
@click.option(
    "--endpoint",
    default="http://127.0.0.1:8669",
)
@click.option(
    "--keystore",
    default="",
)
@click.option(
    "--passcode",
    default="",
)
@click.option(
    "--log",
    default=False,
    type=bool,
)
@click.option(
    "--debug",
    default=False,
    type=bool,
)
def run_server(host, port, endpoint, keystore, passcode, log, debug):
    print('run server')
    try:
        print(endpoint)
        response = requests.options(endpoint)
        response.raise_for_status()
    except requests.exceptions.ConnectionError:
        print("Unable to connect to Meter-Restful server.")
        return

    print(make_version())
    print("Listening on %s:%s" % (host, port))

    meter.set_endpoint(endpoint)
    if keystore == "":
        meter.set_accounts(solo())
    else:
        meter.set_accounts(_keystore(keystore, passcode))

    app = web.Application()
    app.router.add_get("/", lambda r: web.Response(headers=res_headers))
    app.router.add_post("/", lambda r: handle(r, log, debug))
    app.router.add_options("/", lambda r: web.Response(headers=res_headers))
    app.router.add_get(
        "/health", lambda r: web.Response(headers=res_headers, body="OK", content_type="text/plain"))
    web.run_app(app, host=host, port=port)


if __name__ == '__main__':
    run_server()



import rlp
import uuid
from gear.utils.singleton import Singleton
from gear.utils.types import (
    encode_number,
    encode_hex,
    strip_0x
)
from gear.utils.compat import (
    meter_block_convert_to_eth_block,
    meter_receipt_convert_to_eth_receipt,
    meter_tx_convert_to_eth_tx,
    meter_log_convert_to_eth_log,
    meter_storage_convert_to_eth_storage,
    MeterTransaction,
    intrinsic_gas,
)
from .request import (
    Restful,
    get,
    post,
)


def _attribute(obj, key): return None if obj is None else obj[key]


class MeterClient(object, metaclass=Singleton):
    def __init__(self):
        self.filter = {}

    def set_endpoint(self, endpoint):
        restful = Restful(endpoint)
        self.eth_transactions = restful.transactions.eth
        self.transactions = restful.transactions
        self.blocks = restful.blocks
        self.accounts = restful.accounts
        self.events = restful.events
        self.debug = restful.debug

    def set_accounts(self, account_manager):
        self.account_manager = account_manager

    async def trace_transaction(self, tx_hash):
        tx = await self.transactions(tx_hash).make_request(get)
        if tx is None:
            return None
        data = {
            "name": "",
            "target": "{}/{}/0".format(tx["meta"]["blockID"], tx_hash)
        }
        return await self.debug.tracers.make_request(post, data=data)

    async def get_storage_at(self, address, position, block_identifier):
        params = {
            "revision": block_identifier
        }
        storage = await self.accounts(address).storage(
            position).make_request(get, params=params)
        return _attribute(storage, "value")

    async def storage_range_at(self, blk_hash, tx_index, contract_addr, key_start, max_result):
        data = {
            "Address": contract_addr,
            "KeyStart": key_start,
            "MaxResult": max_result,
            "target": "{}/{}/0".format(blk_hash, tx_index)
        }
        result = await self.debug("storage-range").make_request(post, data=data)
        if result is None:
            return None
        result["storage"] = meter_storage_convert_to_eth_storage(
            result["storage"])
        return result

    def get_accounts(self):
        return self.account_manager.get_accounts()

    async def get_block_number(self):
        blk = await self.blocks("best").make_request(get)
        return _attribute(blk, "number")

    async def get_block_id(self, block_identifier):
        blk = await self.blocks(block_identifier).make_request(get)
        return _attribute(blk, "id")

    async def estimate_gas(self, transaction):
        data = {
            "data": transaction["data"],
            "value": (encode_number(transaction.get("value", 0))).decode("utf-8"),
            "caller": transaction.get("from", None),
        }
        result = await self.accounts(transaction.get(
            "to", None)).make_request(post, data=data)
        if result is None or result["reverted"]:
            raise ValueError("Gas estimation failed.")
        return int(result["gasUsed"] * 1.2) + intrinsic_gas(transaction)

    async def call(self, transaction, block_identifier):
        params = {
            "revision": block_identifier,
        }
        data = {
            "data": transaction["data"],
            "value": (encode_number(transaction.get("value", 0))).decode("utf-8"),
            "caller": transaction.get("from", None),
        }
        result = await self.accounts(transaction.get("to", None)).make_request(
            post, data=data, params=params)
        return _attribute(result, "data")

    async def send_transaction(self, transaction):
        chain_tag = int((await meter.get_block(0))["hash"][-2:], 16)
        blk_ref = int(strip_0x((await meter.get_block("best"))["hash"])[:8], 16)
        tx = MeterTransaction(chain_tag, blk_ref, transaction)
        tx.sign(self.account_manager.get_priv_by_addr(transaction["from"]))
        raw = "0x{}".format(encode_hex(rlp.encode(tx)))
        return await self.send_raw_transaction(raw)

    async def send_raw_transaction(self, raw):
        data = {
            "raw": raw
        }
        result = await self.eth_transactions.make_request(post, data=data)
        return _attribute(result, "id")

    async def get_transaction_by_hash(self, tx_hash):
        tx = await self.transactions(tx_hash).make_request(get)
        return None if tx is None else meter_tx_convert_to_eth_tx(tx)

    async def get_balance(self, address, block_identifier):
        params = {
            "revision": block_identifier
        }
        accout = await self.accounts(address).make_request(get, params=params)
        return _attribute(accout, "energy")

    async def get_transaction_receipt(self, tx_hash):
        receipt = await self.transactions(tx_hash).receipt.make_request(get)
        return None if receipt is None else meter_receipt_convert_to_eth_receipt(receipt)

    async def get_block(self, block_identifier):
        blk = await self.blocks(block_identifier).make_request(get)
        return None if blk is None else meter_block_convert_to_eth_block(blk)

    async def get_transaction_count(self, address, tag):
        nonce = '0x1'
        return _attribute(nonce, "count")

    async def get_code(self, address, block_identifier):
        params = {
            "revision": block_identifier
        }
        code = await self.accounts(address).code.make_request(get, params=params)
        return _attribute(code, "code")

    async def new_block_filter(self):
        filter_id = "0x{}".format(uuid.uuid4().hex)
        current_block_num = await self.get_block_number()
        self.filter[filter_id] = BlockFilter(current_block_num, self)
        return filter_id

    def uninstall_filter(self, filter_id):
        if filter_id in self.filter:
            del self.filter[filter_id]
        return True

    async def get_filter_changes(self, filter_id):
        func = self.filter.get(filter_id)
        return await func() if func else []

    async def get_logs(self, address, query):
        params = {
            "address": address
        }
        logs = await self.events.make_request(post, data=query, params=params)
        result = meter_log_convert_to_eth_log(address, logs)
        return result


class BlockFilter(object):

    def __init__(self, current_block_num, client):
        super(BlockFilter, self).__init__()
        self.current = current_block_num
        self.client = client

    async def __call__(self):
        result = []
        best_num = await self.client.get_block_number()
        if best_num:
            result = [
                id
                for id in [
                    await id
                    for id in map(self.client.get_block_id, range(self.current, best_num + 1))
                ]
                if id is not None
            ]
            self.current = best_num + 1
        return result


meter = MeterClient()



import aiohttp


async def post(endpoint_uri, data, **kwargs):
    async with aiohttp.ClientSession() as session:
        return await session.post(endpoint_uri, json=data, **kwargs)


async def get(endpoint_uri, params, **kwargs):
    async with aiohttp.ClientSession() as session:
        return await session.get(endpoint_uri, params=params, **kwargs)


class Restful(object):

    def __init__(self, endpoint):
        super(Restful, self).__init__()
        self._endpoint = endpoint

    def __call__(self, parameter):
        if parameter is not None:
            return Restful('%s/%s' % (self._endpoint, parameter))
        return self

    def __getattr__(self, resource):
        return Restful('%s/%s' % (self._endpoint, resource))

    async def make_request(self, method, params=None, data=None, **kwargs):
        headers = {
            "accept": "application/json",
            "Connection": "keep-alive",
            "Content-Type": "application/json"
        }
        kwargs.setdefault('headers', headers)
        kwargs.setdefault('timeout', 10)
        error = None
        try:
            response = await method(self._endpoint, params=params, data=data, **kwargs)
            return await response.json()
        except aiohttp.ClientConnectionError as e:
            print("Unable to connect to Meter-Restful server:")
            error = e
        except Exception as e:
            try:
                text = await response.text()
                error = Exception(text.strip('\n'))
            except:
                error = e
        print("Meter-Restful server Err:")
        raise error



import itertools
import functools
import logging
import sys
import json
import traceback
from .meter.client import meter
from .utils.compat import noop
from .utils.types import (
    encode_number,
    force_obj_to_text,
    normalize_block_identifier,
    normalize_number
)
from jsonrpcserver import method


def async_serialize(func):
    @functools.wraps(func)
    async def wrapper(*args, **kw):
        try:
            result = await func(*args, **kw)
            if isinstance(result, str):
                return result
            return force_obj_to_text(result, True)
        except Exception as e:
            traceback.print_exc()
            raise e
    return wrapper


#
# formatter
#
TXN_FORMATTERS = {
    'value': normalize_number,
    'gas': normalize_number,
}


def input_transaction_formatter(transaction):
    return {
        k: TXN_FORMATTERS.get(k, noop)(v)
        for k, v in transaction.items()
    }


def input_log_filter_formatter(filter_params):
    params_range = {"unit": "block"}
    params_range["from"] = int(filter_params["fromBlock"], 16)
    to_blk = filter_params.get("toBlock", None)
    if to_blk:
        params_range["to"] = int(to_blk, 16)
    return {
        "range": params_range,
        "topicSets": topics_formatter(filter_params.get("topics", []))
    }


def topics_formatter(eth_topics):
    if eth_topics:
        matrix = [x if isinstance(x, list) else [x] for x in eth_topics]
        return [
            {
                "topic{}".format(index): topic
                for index, topic in enumerate(e)
            }
            for e in itertools.product(*matrix)
        ]
    return []


#
#
#
@method
async def rpc_modules():
    return {
        "eth": "1.0",
        "net": "1.0",
        "web3": "1.0",
    }


#
# debug
#
@method
@async_serialize
async def debug_traceTransaction(tx_hash, params):
    return await meter.trace_transaction(tx_hash)


@method
@async_serialize
async def debug_storageRangeAt(blk_hash, tx_index, contract_addr, key_start, max_result):
    return await meter.storage_range_at(blk_hash, tx_index, contract_addr, key_start, max_result)


#
# net_api
#
@method
async def net_version():
    return 5777


@method
async def net_listening():
    return False

#
# evm_api
# 没有真实实现, 只是为了实现接口
#


@method
async def evm_snapshot():
    return encode_number(0)


@method
async def evm_revert(snapshot_idx=None):
    return True


#
# web3
#
def make_version():
    from . import __version__
    return "Meter-Gear/" + __version__ + "/{platform}/python{v.major}.{v.minor}.{v.micro}".format(
        v=sys.version_info,
        platform=sys.platform,
    )


@method
async def web3_clientVersion():
    return make_version()


#
# eth_api
#
@method
@async_serialize
async def eth_getStorageAt(address, position, block_identifier="best"):
    if position.startswith("0x"):
        position = position[2:]
    position = "0x{}".format(position.zfill(64))
    return await meter.get_storage_at(
        address, position, normalize_block_identifier(block_identifier))


@method
@async_serialize
async def eth_chainId():
    return "0x52"


@method
@async_serialize
async def eth_getTransactionCount(address, block_identifier="best"):
    '''
    ethereum 用来处理 nonce, Meter 不需要
    '''
    return encode_number(0)


@method
async def eth_accounts():
    print('get accounts')
    accounts = meter.get_accounts()
    print(accounts)
    return accounts


@method
@async_serialize
async def eth_getCode(address, block_identifier="best"):
    return await meter.get_code(address, normalize_block_identifier(block_identifier))


@method
@async_serialize
async def eth_blockNumber():
    return encode_number(await meter.get_block_number())


@method
@async_serialize
async def eth_estimateGas(transaction):
    formatted_transaction = input_transaction_formatter(transaction)
    return encode_number(await meter.estimate_gas(formatted_transaction))


@method
@async_serialize
async def eth_call(transaction, block_identifier="best"):
    formatted_transaction = input_transaction_formatter(transaction)
    return await meter.call(formatted_transaction, normalize_block_identifier(block_identifier))


@method
@async_serialize
async def eth_sendTransaction(transaction):
    '''
    发送未签名的交易
    '''
    print('tx: ', transaction)
    formatted_transaction = input_transaction_formatter(transaction)
    return await meter.send_transaction(formatted_transaction)


@method
@async_serialize
async def eth_sign(transaction):
    '''
    发送未签名的交易
    '''
    print('eth_sign tx: ', transaction)
    return await {}


@method
@async_serialize
async def eth_sendRawTransaction(raw):
    '''
    发送已签名的交易
    '''
    print('raw:', raw)
    return await meter.send_raw_transaction(raw)


@method
@async_serialize
async def eth_getBalance(address, block_identifier="best"):
    return await meter.get_balance(address, normalize_block_identifier(block_identifier))


@method
@async_serialize
async def eth_getTransactionByHash(tx_hash):
    if tx_hash:
        return await meter.get_transaction_by_hash(tx_hash)
    return None


@method
@async_serialize
async def eth_getTransactionReceipt(tx_hash):
    if tx_hash:
        return await meter.get_transaction_receipt(tx_hash)
    return None


@method
@async_serialize
async def eth_getBlockByHash(block_hash, full_tx=False):
    return await get_block(block_hash, full_tx)


@method
@async_serialize
async def eth_getBlockByNumber(block_number, full_tx=False):
    return await get_block(block_number, full_tx)


async def get_block(block_identifier, full_tx):
    blk = await meter.get_block(normalize_block_identifier(block_identifier))
    if blk and full_tx:
        blk["transactions"] = [eth_getTransactionByHash(
            tx) for tx in blk["transactions"]]
    return blk


@method
async def eth_newBlockFilter():
    return await meter.new_block_filter()


@method
@async_serialize
async def eth_uninstallFilter(filter_id):
    return meter.uninstall_filter(filter_id)


@method
@async_serialize
async def eth_getFilterChanges(filter_id):
    return await meter.get_filter_changes(filter_id)


@method
@async_serialize
async def eth_getLogs(filter_obj):
    return await meter.get_logs(filter_obj.get("address", None), input_log_filter_formatter(filter_obj))
